@@ -1,4 +1,4 @@
-/* $Id: files.c 4500 2010-04-14 06:03:12Z astyanax $ */
+/* $Id: files.c 4558 2013-01-03 04:50:49Z astyanax $ */
 /**************************************************************************
  *   files.c                                                              *
  *                                                                        *
@@ -31,6 +31,7 @@
 #include <errno.h>
 #include <ctype.h>
 #include <pwd.h>
+#include <libgen.h>
 
 /* Add an entry to the openfile openfilestruct.  This should only be
  * called from open_buffer(). */
@@ -77,6 +78,7 @@ void initialize_buffer(void)
     openfile->current_stat = NULL;
     openfile->undotop = NULL;
     openfile->current_undo = NULL;
+    openfile->lock_filename = NULL;
 #endif
 #ifdef ENABLE_COLOR
     openfile->colorstrings = NULL;
@@ -103,6 +105,207 @@ void initialize_buffer_text(void)
     openfile->totsize = 0;
 }
 
+
+#ifndef NANO_TINY
+/* Actyally write the lock file.  This function will
+   ALWAYS annihilate any previous version of the file.
+   We'll borrow INSECURE_BACKUP here to decide about lock file
+   paranoia here as well...
+   Args:
+       lockfilename: file name for lock
+       origfilename: name of the file the lock is for
+       modified: whether to set the modified bit in the file
+
+   Returns: 1 on success, 0 on failure (but continue loading), -1 on failure and abort
+ */
+int write_lockfile(const char *lockfilename, const char *origfilename, bool modified)
+{
+    int cflags, fd;
+    FILE *filestream;
+    pid_t mypid;
+    uid_t myuid;
+    struct passwd *mypwuid;
+    char *lockdata = charalloc(1024);
+    char myhostname[32];
+    ssize_t lockdatalen = 1024;
+    ssize_t wroteamt;
+
+    /* Run things which might fail first before we try and blow away
+       the old state */
+    myuid = geteuid();
+    if ((mypwuid = getpwuid(myuid)) == NULL) {
+        statusbar(_("Couldn't determine my identity for lock file (getpwuid() failed)"));
+        return -1;
+    }
+    mypid = getpid();
+
+    if (gethostname(myhostname, 31) < 0) {
+       statusbar(_("Couldn't determine hosttname for lock file: %s"), strerror(errno));
+       return -1;
+    }
+
+    if (delete_lockfile(lockfilename) < 0)
+        return -1;
+
+    if (ISSET(INSECURE_BACKUP))
+        cflags = O_WRONLY | O_CREAT | O_APPEND;
+    else
+        cflags = O_WRONLY | O_CREAT | O_EXCL | O_APPEND;
+
+    fd = open(lockfilename, cflags,
+	    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+
+    /* Maybe we just don't have write access, don't stop us from
+       opening the file at all, just don't set the lock_filename
+       and return success */
+    if (fd < 0 && errno == EACCES)
+        return 1;
+
+    /* Now we've got a safe file stream.  If the previous open()
+    call failed, this will return NULL. */
+    filestream = fdopen(fd, "wb");
+
+    if (fd < 0 || filestream == NULL) {
+        statusbar(_("Error writing lock file %s: %s"), lockfilename,
+		    strerror(errno));
+        return -1;
+    }
+
+
+    /* Okay. so at the moment we're following this state for how
+       to store the lock data:
+       byte 0        - 0x62
+       byte 1        - 0x30
+       bytes 2-12    - program name which created the lock
+       bytes 24,25   - little endian store of creator program's PID
+                       (b24 = 256^0 column, b25 = 256^1 column)
+       bytes 28-44   - username of who created the lock
+       bytes 68-100  - hostname of where the lock was created
+       bytes 108-876 - filename the lock is for
+       byte 1007     - 0x55 if file is modified
+
+       Looks like VIM also stores undo state in this file so we're
+       gonna have to figure out how to slap a 'OMG don't use recover
+       on our lockfile' message in here...
+
+       This is likely very wrong, so this is a WIP
+     */
+    null_at(&lockdata, lockdatalen);
+    lockdata[0] = 0x62;
+    lockdata[1] = 0x30;
+    lockdata[24] = mypid % 256;
+    lockdata[25] = mypid / 256;
+    snprintf(&lockdata[2], 10, "nano %s", VERSION);
+    strncpy(&lockdata[28], mypwuid->pw_name, 16);
+    strncpy(&lockdata[68], myhostname, 31);
+    strncpy(&lockdata[108], origfilename, 768);
+    if (modified == TRUE)
+        lockdata[1007] = 0x55;
+
+    wroteamt = fwrite(lockdata, sizeof(char), lockdatalen, filestream);
+    if (wroteamt < lockdatalen) {
+        statusbar(_("Error writing lock file %s: %s"),
+                  lockfilename, ferror(filestream));
+        return -1;
+    }
+
+#ifdef DEBUG
+    fprintf(stderr, "In write_lockfile(), write successful (wrote %d bytes)\n", wroteamt);
+#endif /* DEBUG */
+
+    if (fclose(filestream) == EOF) {
+        statusbar(_("Error writing lock file %s: %s"),
+                  lockfilename, strerror(errno));
+        return -1;
+    }
+
+    openfile->lock_filename = lockfilename;
+
+    return 1;
+}
+
+
+/* Less exciting, delete the lock file.
+   Return -1 if successful and complain on the statusbar, 1 otherwite
+ */
+int delete_lockfile(const char *lockfilename)
+{
+    if (unlink(lockfilename) < 0 && errno != ENOENT) {
+        statusbar(_("Error deleting lock file %s: %s"), lockfilename,
+		    strerror(errno));
+        return -1;
+    }
+    return 1;
+}
+
+
+/* Deal with lockfiles.  Return -1 on refusing to override
+   the lock file, and 1 on successfully created the lockfile, 0 means
+   we were not successful on creating the lockfile but we should
+   continue to load the file and complain to the user.
+ */
+int do_lockfile(const char *filename)
+{
+    char *lockdir = dirname((char *) mallocstrcpy(NULL, filename));
+    char *lockbase = basename((char *) mallocstrcpy(NULL, filename));
+    ssize_t lockfilesize = (sizeof (char *) * (strlen(filename)
+                   + strlen(locking_prefix) + strlen(locking_suffix) + 3));
+    char *lockfilename = (char *) nmalloc(lockfilesize);
+    char lockprog[12], lockuser[16];
+    struct stat fileinfo;
+    int lockfd, lockpid;
+
+
+    snprintf(lockfilename, lockfilesize, "%s/%s%s%s", lockdir,
+             locking_prefix, lockbase, locking_suffix);
+#ifdef DEBUG
+    fprintf(stderr, "lock file name is %s\n", lockfilename);
+#endif /* DEBUG */
+    if (stat(lockfilename, &fileinfo) != -1) {
+        ssize_t readtot = 0;
+        ssize_t readamt = 0;
+        char *lockbuf = (char *) nmalloc(8192);
+        char *promptstr = (char *) nmalloc(128);
+        int ans;
+        if ((lockfd = open(lockfilename, O_RDONLY)) < 0) {
+            statusbar(_("Error opening lockfile %s: %s"),
+                      lockfilename, strerror(errno));
+            return -1;
+        }
+        do {
+            readamt = read(lockfd, &lockbuf[readtot], BUFSIZ);
+            readtot += readamt;
+        } while (readtot < 8192 && readamt > 0);
+
+        if (readtot < 48) {
+            statusbar(_("Error reading lockfile %s: Not enough data read"),
+                      lockfilename);
+            return -1;
+        }
+        strncpy(lockprog, &lockbuf[2], 10);
+        lockpid = lockbuf[25] * 256 + lockbuf[24];
+        strncpy(lockuser, &lockbuf[28], 16);
+#ifdef DEBUG
+        fprintf(stderr, "lockpid = %d\n", lockpid);
+        fprintf(stderr, "program name which created this lock file should be %s\n",
+                lockprog);
+        fprintf(stderr, "user which created this lock file should be %s\n",
+                lockuser);
+#endif /* DEBUG */
+        sprintf(promptstr, "File being edited (by %s, PID %d, user %s), continue?",
+                              lockprog, lockpid, lockuser);
+        ans = do_yesno_prompt(FALSE, promptstr);
+        if (ans < 1) {
+            blank_statusbar();
+            return -1;
+        }
+    }
+
+    return write_lockfile(lockfilename, filename, FALSE);
+}
+#endif /* NANO_TINY */
+
+
 /* If it's not "", filename is a file to open.  We make a new buffer, if
  * necessary, and then open and read the file, if applicable. */
 void open_buffer(const char *filename, bool undoable)
@@ -128,16 +331,16 @@ void open_buffer(const char *filename, bool undoable)
     }
 #endif
 
-    /* If the filename isn't blank, open the file.  Otherwise, treat it
-     * as a new file. */
-    rc = (filename[0] != '\0') ? open_file(filename, new_buffer, &f) :
-	-2;
-
     /* If we're loading into a new buffer, add a new entry to
      * openfile. */
     if (new_buffer)
 	make_new_buffer();
 
+    /* If the filename isn't blank, open the file.  Otherwise, treat it
+     * as a new file. */
+    rc = (filename[0] != '\0') ? open_file(filename, new_buffer, &f) :
+	-2;
+
     /* If we have a file, and we're loading into a new buffer, update
      * the filename. */
     if (rc != -1 && new_buffer)
@@ -279,6 +482,10 @@ bool close_buffer(void)
     if (openfile == openfile->next)
 	return FALSE;
 
+#ifndef NANO_TINY
+        update_poshistory(openfile->filename, openfile->current->lineno, xplustabs()+1);
+#endif /* NANO_TINY */
+
     /* Switch to the next file buffer. */
     switch_to_next_buffer_void();
 
@@ -674,7 +881,7 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable, bool checkw
  * found".
  *
  * Return -2 if we say "New File", -1 if the file isn't opened, and the
- * fd opened otherwise.  The file might still have an error while reading 
+ * fd opened otherwise.  The file might still have an error while reading
  * with a 0 return value.  *f is set to the opened file. */
 int open_file(const char *filename, bool newfie, FILE **f)
 {
@@ -689,10 +896,17 @@ int open_file(const char *filename, bool newfie, FILE **f)
 
     /* Okay, if we can't stat the path due to a component's
        permissions, just try the relative one */
-    if (full_filename == NULL 
+    if (full_filename == NULL
 	|| (stat(full_filename, &fileinfo) == -1 && stat(filename, &fileinfo2) != -1))
 	full_filename = mallocstrcpy(NULL, filename);
 
+
+#ifndef NANO_TINY
+    if (ISSET(LOCKING))
+        if (do_lockfile(full_filename) < 0)
+            return -1;
+#endif
+
     if (stat(full_filename, &fileinfo) == -1) {
 	/* Well, maybe we can open the file even if the OS
 	   says its not there */
@@ -869,14 +1083,14 @@ void do_insertfile(
 #ifndef NANO_TINY
 #ifdef ENABLE_MULTIBUFFER
 
-	    if (s && s->scfunc == NEW_BUFFER_MSG) {
+	    if (s && s->scfunc == new_buffer_void) {
 		/* Don't allow toggling if we're in view mode. */
 		if (!ISSET(VIEW_MODE))
 		    TOGGLE(MULTIBUFFER);
 		continue;
 	    } else
 #endif
-	    if (s && s->scfunc == EXT_CMD_MSG) {
+	    if (s && s->scfunc == ext_cmd_void) {
 		execute = !execute;
 		continue;
 	    }
@@ -886,7 +1100,7 @@ void do_insertfile(
 #endif /* !NANO_TINY */
 
 #ifndef DISABLE_BROWSER
-	    if (s && s->scfunc == TO_FILES_MSG) {
+	    if (s && s->scfunc == to_files_void) {
 		char *tmp = do_browse_from(answer);
 
 		if (tmp == NULL)
@@ -1066,6 +1280,12 @@ void do_insertfile(
  * allow inserting a file into a new buffer. */
 void do_insertfile_void(void)
 {
+
+    if (ISSET(RESTRICTED)) {
+        nano_disabled_msg();
+	return;
+    }
+
 #ifdef ENABLE_MULTIBUFFER
     if (ISSET(VIEW_MODE) && !ISSET(MULTIBUFFER))
 	statusbar(_("Key invalid in non-multibuffer mode"));
@@ -1360,6 +1580,23 @@ bool check_operating_dir(const char *currpath, bool allow_tabcomp)
 #endif
 
 #ifndef NANO_TINY
+/* Although this sucks, it sucks less than having a single 'my system is messed up
+ * and I'm blanket allowing insecure file writing operations.
+ */
+
+int prompt_failed_backupwrite(const char *filename)
+{
+    static int i;
+    static char *prevfile = NULL; /* What was the laast file we were paased so we don't keep asking this?
+                                     though maybe we should.... */
+    if (prevfile == NULL || strcmp(filename, prevfile)) {
+	i = do_yesno_prompt(FALSE,
+                         _("Failed to write backup file, continue saving? (Say N if unsure) "));
+	prevfile = mallocstrcpy(prevfile, filename);
+    }
+    return i;
+}
+
 void init_backup_dir(void)
 {
     char *full_backup_dir;
@@ -1520,6 +1757,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	char *backupname;
 	struct utimbuf filetime;
 	int copy_status;
+	int backup_cflags;
 
 	/* Save the original file's access and modification times. */
 	filetime.actime = openfile->current_stat->st_atime;
@@ -1592,14 +1830,21 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	/* First, unlink any existing backups.  Next, open the backup
 	   file with O_CREAT and O_EXCL.  If it succeeds, we
 	   have a file descriptor to a new backup file. */
-	if (unlink(backupname) < 0 && errno != ENOENT) {
+	if (unlink(backupname) < 0 && errno != ENOENT && !ISSET(INSECURE_BACKUP)) {
+	    if (prompt_failed_backupwrite(backupname))
+		goto skip_backup;
 	    statusbar(_("Error writing backup file %s: %s"), backupname,
 			strerror(errno));
 	    free(backupname);
 	    goto cleanup_and_exit;
 	}
 
-	backup_fd = open(backupname, O_WRONLY | O_CREAT | O_EXCL | O_APPEND,
+	if (ISSET(INSECURE_BACKUP))
+	    backup_cflags = O_WRONLY | O_CREAT | O_APPEND;
+        else
+	    backup_cflags = O_WRONLY | O_CREAT | O_EXCL | O_APPEND;
+
+	backup_fd = open(backupname, backup_cflags,
 		S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
 	/* Now we've got a safe file stream.  If the previous open()
 	   call failed, this will return NULL. */
@@ -1612,9 +1857,24 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	    goto cleanup_and_exit;
 	}
 
-	if (fchmod(backup_fd, openfile->current_stat->st_mode) == -1 ||
-	    fchown(backup_fd, openfile->current_stat->st_uid,
-		   openfile->current_stat->st_gid) == -1 ) {
+        /* We shouldn't worry about chown()ing something if we're not
+	   root, since it's likely to fail! */
+	if (geteuid() == NANO_ROOT_UID && fchown(backup_fd,
+		openfile->current_stat->st_uid, openfile->current_stat->st_gid) == -1
+		&& !ISSET(INSECURE_BACKUP)) {
+	    if (prompt_failed_backupwrite(backupname))
+		goto skip_backup;
+	    statusbar(_("Error writing backup file %s: %s"), backupname,
+		strerror(errno));
+	    free(backupname);
+	    fclose(backup_file);
+	    goto cleanup_and_exit;
+	}
+
+	if (fchmod(backup_fd, openfile->current_stat->st_mode) == -1
+		&& !ISSET(INSECURE_BACKUP)) {
+	    if (prompt_failed_backupwrite(backupname))
+		goto skip_backup;
 	    statusbar(_("Error writing backup file %s: %s"), backupname,
 		strerror(errno));
 	    free(backupname);
@@ -1633,14 +1893,18 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	/* Copy the file. */
 	copy_status = copy_file(f, backup_file);
 
-	/* And set its metadata. */
-	if (copy_status != 0  || utime(backupname, &filetime) == -1) {
-	    if (copy_status == -1) {
-		statusbar(_("Error reading %s: %s"), realname,
+	if (copy_status != 0) {
+	    statusbar(_("Error reading %s: %s"), realname,
 			strerror(errno));
-		beep();
-	    } else
-		statusbar(_("Error writing backup file %s: %s"), backupname,
+	    beep();
+	    goto cleanup_and_exit;
+	}
+
+	/* And set its metadata. */
+	if (utime(backupname, &filetime) == -1 && !ISSET(INSECURE_BACKUP)) {
+	    if (prompt_failed_backupwrite(backupname))
+		goto skip_backup;
+	    statusbar(_("Error writing backup file %s: %s"), backupname,
 			strerror(errno));
 	    /* If we can't write to the backup, DONT go on, since
 	       whatever caused the backup file to fail (e.g. disk
@@ -1936,6 +2200,7 @@ bool write_marked_file(const char *name, FILE *f_open, bool tmp,
 
     return retval;
 }
+
 #endif /* !NANO_TINY */
 
 /* Write the current file to disk.  If the mark is on, write the current
@@ -2029,7 +2294,7 @@ bool do_writeout(bool exiting)
             s = get_shortcut(currmenu, &i, &meta_key, &func_key);
 
 #ifndef DISABLE_BROWSER
-	    if (s && s->scfunc == TO_FILES_MSG) {
+	    if (s && s->scfunc == to_files_void) {
 		char *tmp = do_browse_from(answer);
 
 		if (tmp == NULL)
@@ -2041,26 +2306,26 @@ bool do_writeout(bool exiting)
 	    } else
 #endif /* !DISABLE_BROWSER */
 #ifndef NANO_TINY
-	    if (s && s->scfunc ==  DOS_FORMAT_MSG) {
+	    if (s && s->scfunc == dos_format_void) {
 		openfile->fmt = (openfile->fmt == DOS_FILE) ? NIX_FILE :
 			DOS_FILE;
 		continue;
-	    } else if (s && s->scfunc ==  MAC_FORMAT_MSG) {
+	    } else if (s && s->scfunc == mac_format_void) {
 		openfile->fmt = (openfile->fmt == MAC_FILE) ? NIX_FILE :
 			MAC_FILE;
 		continue;
-	    } else if (s && s->scfunc ==  BACKUP_FILE_MSG) {
+	    } else if (s && s->scfunc == backup_file_void) {
 		TOGGLE(BACKUP_FILE);
 		continue;
 	    } else
 #endif /* !NANO_TINY */
-	    if (s && s->scfunc ==  PREPEND_MSG) {
+	    if (s && s->scfunc == prepend_void) {
 		append = (append == PREPEND) ? OVERWRITE : PREPEND;
 		continue;
-	    } else if (s && s->scfunc ==  APPEND_MSG) {
+	    } else if (s && s->scfunc == append_void) {
 		append = (append == APPEND) ? OVERWRITE : APPEND;
 		continue;
-	    } else if (s && s->scfunc == DO_HELP_VOID) {
+	    } else if (s && s->scfunc == do_help_void) {
 		continue;
 	    }
 
@@ -2348,9 +2613,6 @@ char **cwd_tab_completion(const char *buf, bool allow_files, size_t
 	*num_matches, size_t buf_len)
 {
     char *dirname = mallocstrcpy(NULL, buf), *filename;
-#ifndef DISABLE_OPERATINGDIR
-    size_t dirnamelen;
-#endif
     size_t filenamelen;
     char **matches = NULL;
     DIR *dir;
@@ -2389,9 +2651,6 @@ char **cwd_tab_completion(const char *buf, bool allow_files, size_t
 	return NULL;
     }
 
-#ifndef DISABLE_OPERATINGDIR
-    dirnamelen = strlen(dirname);
-#endif
     filenamelen = strlen(filename);
 
     while ((nextdir = readdir(dir)) != NULL) {
@@ -2633,28 +2892,99 @@ const char *tail(const char *foo)
 }
 
 #if !defined(NANO_TINY) && defined(ENABLE_NANORC)
-/* Return $HOME/.nano_history, or NULL if we can't find the home
+/* Return the constructed dorfile path, or NULL if we can't find the home
  * directory.  The string is dynamically allocated, and should be
  * freed. */
-char *histfilename(void)
+char *construct_filename(const char *str)
 {
-    char *nanohist = NULL;
+    char *newstr = NULL;
 
     if (homedir != NULL) {
 	size_t homelen = strlen(homedir);
 
-	nanohist = charalloc(homelen + 15);
-	strcpy(nanohist, homedir);
-	strcpy(nanohist + homelen, "/.nano_history");
+	newstr = charalloc(homelen + strlen(str) + 1);
+	strcpy(newstr, homedir);
+	strcpy(newstr + homelen, str);
     }
 
-    return nanohist;
+    return newstr;
+
+}
+
+char *histfilename(void)
+{
+    return construct_filename("/.nano/search_history");
+}
+
+/* Construct the legacy history filename
+ * (Deprecate in 2.5, delete later
+ */
+char *legacyhistfilename(void)
+{
+    return construct_filename("/.nano_history");
+}
+
+char *poshistfilename(void)
+{
+    return construct_filename("/.nano/filepos_history");
+}
+
+
+
+void history_error(const char *msg, ...)
+{
+   va_list ap;
+
+    va_start(ap, msg);
+    vfprintf(stderr, _(msg), ap);
+    va_end(ap);
+
+    fprintf(stderr, _("\nPress Enter to continue\n"));
+    while (getchar() != '\n')
+	;
+
+}
+
+/* Now that we have more than one history file, let's just rely
+   on a .nano dir for this stuff.  Return 1 if the dir exists
+   or was successfully created, and return 0 otherwise.
+ */
+int check_dotnano(void)
+{
+    struct stat dirstat;
+    char *nanodir = construct_filename("/.nano");
+
+    if (stat(nanodir, &dirstat) == -1) {
+	if (mkdir(nanodir, S_IRWXU | S_IRWXG | S_IRWXO) == -1) {
+	    history_error(N_("Unable to create directory %s: %s\nIt is required for saving/loading search history or cursor position\n"),
+		nanodir, strerror(errno));
+	    return 0;
+	}
+    } else if (!S_ISDIR(dirstat.st_mode)) {
+	history_error(N_("Path %s is not a directory and needs to be.\nNano will be unable to load or save search or cursor position history\n"));
+	return 0;
+    }
+    return 1;
 }
 
 /* Load histories from ~/.nano_history. */
 void load_history(void)
 {
     char *nanohist = histfilename();
+    char *legacyhist = legacyhistfilename();
+    struct stat hstat;
+
+
+    if (stat(legacyhist, &hstat) != -1 && stat(nanohist, &hstat) == -1) {
+	if (rename(legacyhist, nanohist)  == -1)
+	    history_error(N_("Detected a legacy nano history file (%s) which I tried to move\nto the preferred location (%s) but encountered an error: %s"),
+		legacyhist, nanohist, strerror(errno));
+	else
+	    history_error(N_("Detected a legacy nano history file (%s) which I moved\nto the preferred location (%s)\n(see the nano FAQ about this change)"),
+		legacyhist, nanohist);
+    }
+
+
 
     /* Assume do_rcfile() has reported a missing home directory. */
     if (nanohist != NULL) {
@@ -2664,12 +2994,8 @@ void load_history(void)
 	    if (errno != ENOENT) {
 		/* Don't save history when we quit. */
 		UNSET(HISTORYLOG);
-		rcfile_error(N_("Error reading %s: %s"), nanohist,
+		history_error(N_("Error reading %s: %s"), nanohist,
 			strerror(errno));
-		fprintf(stderr,
-			_("\nPress Enter to continue starting nano.\n"));
-		while (getchar() != '\n')
-		    ;
 	    }
 	} else {
 	    /* Load a history list (first the search history, then the
@@ -2694,8 +3020,11 @@ void load_history(void)
 
 	    fclose(hist);
 	    free(line);
+	    if (search_history->prev != NULL)
+		last_search = mallocstrcpy(NULL, search_history->prev->data);
 	}
 	free(nanohist);
+	free(legacyhist);
     }
 }
 
@@ -2721,7 +3050,7 @@ bool writehist(FILE *hist, filestruct *h)
     return TRUE;
 }
 
-/* Save histories to ~/.nano_history. */
+/* Save histories to ~/.nano/search_history. */
 void save_history(void)
 {
     char *nanohist;
@@ -2737,7 +3066,7 @@ void save_history(void)
 	FILE *hist = fopen(nanohist, "wb");
 
 	if (hist == NULL)
-	    rcfile_error(N_("Error writing %s: %s"), nanohist,
+	    history_error(N_("Error writing %s: %s"), nanohist,
 		strerror(errno));
 	else {
 	    /* Make sure no one else can read from or write to the
@@ -2746,7 +3075,7 @@ void save_history(void)
 
 	    if (!writehist(hist, searchage) || !writehist(hist,
 		replaceage))
-		rcfile_error(N_("Error writing %s: %s"), nanohist,
+		history_error(N_("Error writing %s: %s"), nanohist,
 			strerror(errno));
 
 	    fclose(hist);
@@ -2755,4 +3084,163 @@ void save_history(void)
 	free(nanohist);
     }
 }
+
+
+/* Analogs for the POS history */
+void save_poshistory(void)
+{
+    char *poshist;
+    char *statusstr = NULL;
+    poshiststruct *posptr;
+
+    poshist = poshistfilename();
+
+    if (poshist != NULL) {
+	FILE *hist = fopen(poshist, "wb");
+
+	if (hist == NULL)
+	    history_error(N_("Error writing %s: %s"), poshist,
+		strerror(errno));
+	else {
+	    /* Make sure no one else can read from or write to the
+	     * history file. */
+	    chmod(poshist, S_IRUSR | S_IWUSR);
+
+            for (posptr = poshistory; posptr != NULL; posptr = posptr->next) {
+		statusstr = charalloc(strlen(posptr->filename) + 2 * sizeof(ssize_t) + 4);
+		sprintf(statusstr, "%s %d %d\n", posptr->filename, (int) posptr->lineno,
+			(int) posptr->xno);
+		if (fwrite(statusstr, sizeof(char), strlen(statusstr), hist) < strlen(statusstr))
+		    history_error(N_("Error writing %s: %s"), poshist,
+			strerror(errno));
+		free(statusstr);
+	    }
+	    fclose(hist);
+	}
+	free(poshist);
+    }
+}
+
+/* Update the POS history, given a filename line and column.
+ * If no entry is found, add a new entry on the end
+ */
+void update_poshistory(char *filename, ssize_t lineno, ssize_t xpos)
+{
+   poshiststruct *posptr, *posprev = NULL;
+   char *fullpath = get_full_path(filename);
+
+    if (fullpath == NULL)
+        return;
+
+    for (posptr = poshistory; posptr != NULL; posptr = posptr->next) {
+        if (!strcmp(posptr->filename, fullpath)) {
+	    posptr->lineno = lineno;
+	    posptr->xno = xpos;
+            return;
+        }
+	posprev = posptr;
+    }
+
+    /* Didn't find it, make a new node yo! */
+
+    posptr = (poshiststruct *) nmalloc(sizeof(poshiststruct));
+    posptr->filename = mallocstrcpy(NULL, fullpath);
+    posptr->lineno = lineno;
+    posptr->xno = xpos;
+    posptr->next = NULL;
+
+    if (!poshistory)
+	poshistory = posptr;
+    else
+	posprev->next = posptr;
+
+    free(fullpath);
+}
+
+
+/* Check the POS history to see if file matches
+ * an existing entry.  If so return 1 and set line and column
+ * to the right values  Otherwise return 0
+ */
+int check_poshistory(const char *file, ssize_t *line, ssize_t *column)
+{
+    poshiststruct *posptr;
+    char *fullpath = get_full_path(file);
+
+    if (fullpath == NULL)
+    	return 0;
+
+    for (posptr = poshistory; posptr != NULL; posptr = posptr->next) {
+	if (!strcmp(posptr->filename, fullpath)) {
+	    *line = posptr->lineno;
+	    *column = posptr->xno;
+	    free(fullpath);
+	    return 1;
+	}
+    }
+    free(fullpath);
+    return 0;
+}
+
+
+/* Load histories from ~/.nano_history. */
+void load_poshistory(void)
+{
+    char *nanohist = poshistfilename();
+
+    /* Assume do_rcfile() has reported a missing home directory. */
+    if (nanohist != NULL) {
+	FILE *hist = fopen(nanohist, "rb");
+
+	if (hist == NULL) {
+	    if (errno != ENOENT) {
+		/* Don't save history when we quit. */
+		UNSET(HISTORYLOG);
+		history_error(N_("Error reading %s: %s"), nanohist,
+			strerror(errno));
+	    }
+	} else {
+	    char *line = NULL, *lineptr, *xptr;
+	    size_t buf_len = 0;
+	    ssize_t read, lineno, xno;
+	    poshiststruct *posptr;
+
+	    /* See if we can find the file we're currently editing */
+	    while ((read = getline(&line, &buf_len, hist)) >= 0) {
+		if (read > 0 && line[read - 1] == '\n') {
+		    read--;
+		    line[read] = '\0';
+		}
+		if (read > 0) {
+		    unsunder(line, read);
+		}
+		lineptr = parse_next_word(line);
+		xptr = parse_next_word(lineptr);
+		lineno = atoi(lineptr);
+		xno = atoi(xptr);
+		if (poshistory == NULL) {
+		    poshistory = (poshiststruct *) nmalloc(sizeof(poshiststruct));
+		    poshistory->filename = mallocstrcpy(NULL, line);
+		    poshistory->lineno = lineno;
+		    poshistory->xno = xno;
+		    poshistory->next = NULL;
+		} else {
+		    for (posptr = poshistory; posptr->next != NULL; posptr = posptr->next)
+			;
+		    posptr->next = (poshiststruct *) nmalloc(sizeof(poshiststruct));
+		    posptr->next->filename = mallocstrcpy(NULL, line);
+		    posptr->next->lineno = lineno;
+		    posptr->next->xno = xno;
+		    posptr->next->next = NULL;
+		}
+
+	    }
+
+	    fclose(hist);
+	    free(line);
+	}
+	free(nanohist);
+    }
+}
+
 #endif /* !NANO_TINY && ENABLE_NANORC */
